#include <stdint.h>
#include <stdio.h>


#define NUM_ITER (1000)
#define NUM_INNER_ITER (1000*1000)


static inline uint64_t rdtsc(void)
{
  uint32_t eax, edx;
  asm volatile ("rdtsc" : "=a"(eax), "=d"(edx));
  return eax + ((uint64_t)edx << 32);
}


int main(void)
{
  uint64_t start_cycles, end_cycles;
  uint32_t i;
  uint32_t j;
#ifdef __x86_64__
  const char *arch = "x86-64";
#elif defined(__i386__)
  const char *arch = "i386";
#endif

#if TEST==1 || !defined(TEST)
  /* Test loop without VZEROUPPER. */
  asm volatile (
    ".align 64\n"
    "pcmpeqd %%xmm0, %%xmm0\n"
    :
    :
    : "memory"
  );
  j = 0; i = 0;
  start_cycles = rdtsc();
    asm volatile (
      ".align 64\n1:\n"
      "pcmpeqd %%xmm7, %%xmm7\n"

      "xor %0, %0\n"
      ".align 64\n2:\n"
      "movdqa %%xmm0, %%xmm1\n"
      "paddd %%xmm1, %%xmm0\n"
      "lea 1(%0),%0\n"
      "cmp %2,%0\n"
      "jb 2b\n"

      "lea 1(%1),%1\n"
      "cmp %3,%1\n"
      "jb 1b\n"
      : "+r" (j), "+r" (i)
      : "r" (NUM_INNER_ITER), "r" (NUM_ITER)
      : "memory", "cc"
    );
  end_cycles = rdtsc();
  printf("%6s-test1: %ld*%ld iterations, %lld cycles\n", arch,
	 (long)i, (long)j, (long long int)(end_cycles - start_cycles));
#endif


#if TEST==2 || !defined(TEST)
  /* Test loop with VZEROUPPER. */
  asm volatile (
    ".align 64\n"
    "pcmpeqd %%xmm0, %%xmm0\n"
    :
    :
    : "memory"
  );
  j = 0; i = 0;
  start_cycles = rdtsc();
    asm volatile (
      ".align 64\n1:\n"
      "vzeroupper\n"

      "xor %0, %0\n"
      ".align 64\n2:\n"
      "movdqa %%xmm0, %%xmm1\n"
      "paddd %%xmm1, %%xmm0\n"
      "lea 1(%0),%0\n"
      "cmp %2,%0\n"
      "jb 2b\n"

      "lea 1(%1),%1\n"
      "cmp %3,%1\n"
      "jb 1b\n"
      : "+r" (j), "+r" (i)
      : "r" (NUM_INNER_ITER), "r" (NUM_ITER)
      : "memory", "cc"
    );
  end_cycles = rdtsc();
  printf("%6s-test2: %ld*%ld iterations, %lld cycles\n", arch,
	 (long)i, (long)j, (long long int)(end_cycles - start_cycles));
#endif


#if TEST==3 || !defined(TEST)
  /* Test loop with VZEROUPPER & YMM. */
  asm volatile (
    ".align 64\n"
    "pcmpeqd %%xmm0, %%xmm0\n"
    :
    :
    : "memory"
  );
  j = 0; i = 0;
  start_cycles = rdtsc();
    asm volatile (
      ".align 64\n1:\n"
      "vmovdqa %%ymm0, %%ymm1\n"
      "vzeroupper\n"

      "xor %0, %0\n"
      ".align 64\n2:\n"
      "movdqa %%xmm0, %%xmm1\n"
      "paddd %%xmm1, %%xmm0\n"
      "lea 1(%0),%0\n"
      "cmp %2,%0\n"
      "jb 2b\n"

      "lea 1(%1),%1\n"
      "cmp %3,%1\n"
      "jb 1b\n"
      : "+r" (j), "+r" (i)
      : "r" (NUM_INNER_ITER), "r" (NUM_ITER)
      : "memory", "cc"
    );
  end_cycles = rdtsc();
  printf("%6s-test3: %ld*%ld iterations, %lld cycles\n", arch,
	 (long)i, (long)j, (long long int)(end_cycles - start_cycles));
#endif


#if TEST==4 || !defined(TEST)
  /* Test loop with VZEROUPPER & YMM & VPXOR. */
  asm volatile (
    ".align 64\n"
    "pcmpeqd %%xmm0, %%xmm0\n"
    :
    :
    : "memory"
  );
  j = 0; i = 0;
  start_cycles = rdtsc();
    asm volatile (
      ".align 64\n1:\n"
      "vmovdqa %%ymm0, %%ymm1\n"
      "vzeroupper\n"
      "vpxor %%xmm1, %%xmm1, %%xmm1\n"

      "xor %0, %0\n"
      ".align 64\n2:\n"
      "movdqa %%xmm0, %%xmm1\n"
      "paddd %%xmm1, %%xmm0\n"
      "lea 1(%0),%0\n"
      "cmp %2,%0\n"
      "jb 2b\n"

      "lea 1(%1),%1\n"
      "cmp %3,%1\n"
      "jb 1b\n"
      : "+r" (j), "+r" (i)
      : "r" (NUM_INNER_ITER), "r" (NUM_ITER)
      : "memory", "cc"
    );
  end_cycles = rdtsc();
  printf("%6s-test4: %ld*%ld iterations, %lld cycles\n", arch,
	 (long)i, (long)j, (long long int)(end_cycles - start_cycles));
#endif


#if TEST==5 || !defined(TEST)
  /* Test loop with VZEROALL. */
  asm volatile (
    ".align 64\n"
    "pcmpeqd %%xmm0, %%xmm0\n"
    :
    :
    : "memory"
  );
  j = 0; i = 0;
  start_cycles = rdtsc();
    asm volatile (
      "xor %0, %0\n"
      ".align 64\n1:\n"
      "vzeroall\n"

      "xor %0, %0\n"
      ".align 64\n2:\n"
      "movdqa %%xmm0, %%xmm1\n"
      "paddd %%xmm1, %%xmm0\n"
      "lea 1(%0),%0\n"
      "cmp %2,%0\n"
      "jb 2b\n"

      "lea 1(%1),%1\n"
      "cmp %3,%1\n"
      "jb 1b\n"
      : "+r" (j), "+r" (i)
      : "r" (NUM_INNER_ITER), "r" (NUM_ITER)
      : "memory", "cc"
    );
  end_cycles = rdtsc();
  printf("%6s-test5: %ld*%ld iterations, %lld cycles\n", arch,
	 (long)i, (long)j, (long long int)(end_cycles - start_cycles));
#endif


#if TEST==6 || !defined(TEST)
  /* Test loop with VZEROUPPER & YMM, VEX in inner loop. */
  asm volatile (
    ".align 64\n"
    "pcmpeqd %%xmm0, %%xmm0\n"
    :
    :
    : "memory"
  );
  j = 0; i = 0;
  start_cycles = rdtsc();
    asm volatile (
      "xor %0, %0\n"
      ".align 64\n1:\n"
      "vmovdqa %%ymm0, %%ymm1\n"
      "vzeroupper\n"

      "xor %0, %0\n"
      ".align 64\n2:\n"
      "vmovdqa %%xmm0, %%xmm1\n"
      "vpaddd %%xmm1, %%xmm0, %%xmm0\n"
      "lea 1(%0),%0\n"
      "cmp %2,%0\n"
      "jb 2b\n"

      "lea 1(%1),%1\n"
      "cmp %3,%1\n"
      "jb 1b\n"
      : "+r" (j), "+r" (i)
      : "r" (NUM_INNER_ITER), "r" (NUM_ITER)
      : "memory", "cc"
    );
  end_cycles = rdtsc();
  printf("%6s-test6: %ld*%ld iterations, %lld cycles\n", arch,
	 (long)i, (long)j, (long long int)(end_cycles - start_cycles));
#endif


  return 0;
}
